/*
    Copyright (C) 2005-2012  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <cairo.h>
#include <pango/pangocairo.h>

#include "gui.h"
#include "nntpgrab_utils.h"

static int bytes_received[10];
static time_t stamp = 0;
static double average_speed = 0;

#if !GTK_CHECK_VERSION(2,14,0)
#define gtk_widget_get_window(x) x->window
#endif

#if GTK_CHECK_VERSION(2,90,0)
static gboolean
expose_event_callback (GtkWidget *widget, cairo_t *cr, gpointer user_data)
#else
static gboolean
expose_event_callback (GtkWidget *widget, GdkEventExpose *event, gpointer data)
#endif
{
    static PangoFontDescription *font_description = NULL;
    PangoLayout *layout;
#if !GTK_CHECK_VERSION(2,90,0)
    cairo_t *cr;
#endif
    int i;
    int max = 0;
    int width;
    int height;
    double y_factor;
    int y_offset;
    double x, y;
    char msg[64];
    char avg_speed[64];
    int text_width;

#if GTK_CHECK_VERSION(2,18,0)
    GtkAllocation allocation;

#if !GTK_CHECK_VERSION(2,90,0)
    cr = gdk_cairo_create(GDK_DRAWABLE(gtk_widget_get_window(widget)));
#endif

    gtk_widget_get_allocation(widget, &allocation);
    width = allocation.width;
    height = allocation.height - (allocation.height / 10);  // Take 10% off the top
    y_offset = allocation.height / 10;  // 10%
#else
    cr = gdk_cairo_create(GDK_DRAWABLE(widget->window));

    width = widget->allocation.width;
    height = widget->allocation.height - (widget->allocation.height / 10);  // Take 10% off the top
    y_offset = widget->allocation.height / 10;  // 10%
#endif

    /* If this is the first time we got here, initialize the font */
    if (!font_description) {
        font_description = pango_font_description_new ();
        //pango_font_description_set_family (font_description, "serif");
        //pango_font_description_set_weight (font_description, PANGO_WEIGHT_BOLD);
        pango_font_description_set_absolute_size (font_description, 12 * PANGO_SCALE);
    }

    // Calculate the peak value
    for (i = 0; i < 10; i++) {
        if (bytes_received[i] > max) {
            max = bytes_received[i];
        }
    }

    // Draw the background
    cairo_set_line_width (cr, 1.0);
    cairo_set_source_rgb (cr, 255, 255, 255);
#if GTK_CHECK_VERSION(2,18,0)
    cairo_rectangle (cr, 0, 0, width, allocation.height);
#else
    cairo_rectangle (cr, 0, 0, width, widget->allocation.height);
#endif
    cairo_fill (cr);

    if (max == 0) {
        // Nothing to draw yet
#if !GTK_CHECK_VERSION(2,90,0)
        cairo_destroy(cr);
#endif
        return TRUE;
    }

    y_factor = height;
    y_factor /= max;

    cairo_set_source_rgb (cr, 0.565, 0.7, 0.87);

    // Create a new path in which all the coordinates are defined
    cairo_new_path(cr);

    // Use the first coordinate as starting point
    y = bytes_received[0];
    y *= y_factor;
    y = height - y;
    y += y_offset;

    cairo_move_to (cr, 0, y);

    // Create a path through the rest of the coordinates
    for (i = 1; i < 10; i++) {
        x = width;
        x /= 9;
        x *= i;

        y = bytes_received[i];
        y *= y_factor;
        y = height - y;
        y += y_offset;

        cairo_line_to (cr, x, y);
    }

    // Finalize the path by going to the bottom right corner
    cairo_line_to(cr, width, height + y_offset);

    // and to the bottom right corner
    cairo_line_to(cr, 0, height + y_offset);

    // Connect back to the first coordinate
    cairo_close_path(cr);

    // Fill the path
    cairo_set_source_rgb (cr, 0.565, 0.7, 0.87);
    cairo_fill (cr);

    // Draw the average speed
    memset(&avg_speed, 0, sizeof(avg_speed));
    memset(&msg, 0, sizeof(msg));
    nntpgrab_utils_calculate_file_size((guint64) average_speed, avg_speed, sizeof(avg_speed) - 1);
    snprintf(msg, sizeof(msg) - 1, "%s/sec", avg_speed);

    layout = pango_cairo_create_layout (cr);
    pango_layout_set_font_description (layout, font_description);
    pango_layout_set_text (layout, msg, -1);

    cairo_set_source_rgb (cr, 0.0, 0.0, 0.0);
    text_width = 0;
    pango_layout_get_pixel_size(layout, &text_width, NULL);
    cairo_move_to (cr, width - text_width - 5, height - 5);
    pango_cairo_show_layout (cr, layout);

    g_object_unref (layout);

#if !GTK_CHECK_VERSION(2,90,0)
    cairo_destroy(cr);
#endif

    return TRUE;
}

static void
traffic_monitor_update_cb (NntpgrabGlue *obj, int bytes_received1, int bytes_received2, int bytes_received3, int bytes_received4, int bytes_received5, int bytes_received6, int bytes_received7, int bytes_received8, int bytes_received9, int bytes_received10, guint64 last_update_stamp, double average)
{
    static GtkWidget *windowMain = NULL;
    static GtkWidget *drawingareaSpeed = NULL;

    if (!windowMain) {
        windowMain = nntpgrab_gui_base_get_widget("windowMain");
    }

    if (!drawingareaSpeed) {
        drawingareaSpeed = nntpgrab_gui_base_get_widget("drawingareaSpeed");
    }

    bytes_received[0] = bytes_received1;
    bytes_received[1] = bytes_received2;
    bytes_received[2] = bytes_received3;
    bytes_received[3] = bytes_received4;
    bytes_received[4] = bytes_received5;
    bytes_received[5] = bytes_received6;
    bytes_received[6] = bytes_received7;
    bytes_received[7] = bytes_received8;
    bytes_received[8] = bytes_received9;
    bytes_received[9] = bytes_received10;

    stamp = last_update_stamp;
    average_speed = average;

    if (gdk_window_get_state(gtk_widget_get_window(windowMain)) != GDK_WINDOW_STATE_ICONIFIED) {
        gdk_window_invalidate_rect(gtk_widget_get_window(drawingareaSpeed), NULL, FALSE);
    }
}

void
monitor_initialize(void)
{
    GtkWidget *drawingareaSpeed = nntpgrab_gui_base_get_widget("drawingareaSpeed");

#if GTK_CHECK_VERSION(2,90,0)
    g_signal_connect (G_OBJECT (drawingareaSpeed), "draw", G_CALLBACK(expose_event_callback), NULL);
#else
    g_signal_connect (G_OBJECT (drawingareaSpeed), "expose_event", G_CALLBACK (expose_event_callback), NULL);
#endif

    nntpgrab_glue_signal_connect(glue, "traffic_monitor_update", NG_CALLBACK(traffic_monitor_update_cb), NULL);
}
