#include <QFileDialog>
#include <QMouseEvent>
#include <QXmlStreamReader>
#include <QStyledItemDelegate>

#include "DialogImportNZB.h"
#include "ui_DialogImportNZB.h"
#include "DialogSetCollectionName.h"

enum {
    FIELD_DOWNLOAD,
    FIELD_SUBJECT,
    FIELD_POSTER,
    FIELD_GROUPS,
    FIELD_STAMP,
    FIELD_FILE_SIZE_STR,
    FIELD_FILE_SIZE,
    FIELD_GROUPS_LIST,
    FIELD_SEGMENTS_LIST,
    NUM_FIELDS
};

class ItemDelegate : public QStyledItemDelegate
{
public:
    ItemDelegate(QObject *parent = 0)
        : QStyledItemDelegate(parent)
    {
    }

    void paint ( QPainter * painter, const QStyleOptionViewItem & option, const QModelIndex & index ) const
    {
        QStyleOptionViewItemV4 viewItemOption(option);

        if (index.column() == 0) {
            const int textMargin = QApplication::style()->pixelMetric(QStyle::PM_FocusFrameHMargin) + 1;
            QRect newRect = QStyle::alignedRect(option.direction, Qt::AlignCenter,
                                                QSize(option.decorationSize.width() + 5,option.decorationSize.height()),
                                                QRect(option.rect.x() + textMargin, option.rect.y(),
                                                      option.rect.width() - (2 * textMargin), option.rect.height()));
            viewItemOption.rect = newRect;
        }
        QStyledItemDelegate::paint(painter, viewItemOption, index);
    }

    virtual bool editorEvent(QEvent *event, QAbstractItemModel *model, const QStyleOptionViewItem &option,
                             const QModelIndex &index)
    {
        Q_ASSERT(event);
        Q_ASSERT(model);

        // make sure that the item is checkable
        Qt::ItemFlags flags = model->flags(index);
        if (!(flags & Qt::ItemIsUserCheckable) || !(flags & Qt::ItemIsEnabled))
            return false;
        // make sure that we have a check state
        QVariant value = index.data(Qt::CheckStateRole);

        // make sure that we have the right event type
        if (event->type() == QEvent::MouseButtonRelease) {
            const int textMargin = QApplication::style()->pixelMetric(QStyle::PM_FocusFrameHMargin) + 1;
            QRect checkRect = QStyle::alignedRect(option.direction, Qt::AlignCenter,
                                                  option.decorationSize,
                                                  QRect(option.rect.x() + (2 * textMargin), option.rect.y(),
                                                        option.rect.width() - (2 * textMargin),
                                                        option.rect.height()));
            if (!checkRect.contains(static_cast<QMouseEvent*>(event)->pos()))
                return false;
        } else if (event->type() == QEvent::KeyPress) {
            if (static_cast<QKeyEvent*>(event)->key() != Qt::Key_Space&& static_cast<QKeyEvent*>(event)->key() != Qt::Key_Select)
                return false;

            QTreeWidget *tree = static_cast<QTreeWidget*>(model->parent());

            for (int i = 0; i < tree->topLevelItemCount(); i++) {
                QTreeWidgetItem *item = tree->topLevelItem(i);

                if(item->isSelected()) {
                    Qt::CheckState checkState = (item->checkState(FIELD_DOWNLOAD) == Qt::Checked ? Qt::Unchecked : Qt::Checked);
                    item->setCheckState(FIELD_DOWNLOAD, checkState);
                }
            }

            return false;
        } else {
            return false;
        }

        if (!value.isValid()) {
            return false;
        }

        Qt::CheckState state = (static_cast<Qt::CheckState>(value.toInt()) == Qt::Checked
                                ? Qt::Unchecked : Qt::Checked);
        return model->setData(index, state, Qt::CheckStateRole);
    }
};

Q_DECLARE_METATYPE(QList<QString>);
Q_DECLARE_METATYPE(QList<QNNTPGrabGlue::Part>);

DialogImportNZB::DialogImportNZB(QNNTPGrabGlue *glue, QWidget *parent) :
    QDialog(parent),
    ui(new Ui::DialogImportNZB)
{
    this->glue = glue;

    ui->setupUi(this);

    connect(ui->btnSelectNZBFile, SIGNAL(clicked()), SLOT(btnSelectNZBFile_clicked()));
    connect(ui->btnSelectAll, SIGNAL(clicked()), SLOT(btnSelectAll_clicked()));
    connect(ui->btnSelectNone, SIGNAL(clicked()), SLOT(btnSelectNone_clicked()));
    connect(ui->btnCancel, SIGNAL(clicked()), SLOT(btnCancel_clicked()));
    connect(ui->btnAdd, SIGNAL(clicked()), SLOT(btnAdd_clicked()));

    /* Disable the 'Add' button by default as the user needs to select a NZB file first */
    ui->btnAdd->setEnabled(false);

    /* Add a custom delegate to draw the checkbuttons centered */
    ui->treeWidget->setItemDelegate(new ItemDelegate(this));

    ui->treeWidget->setColumnCount(NUM_FIELDS);
    ui->treeWidget->setColumnHidden(FIELD_FILE_SIZE, true);
    ui->treeWidget->setColumnHidden(FIELD_GROUPS_LIST, true);
    ui->treeWidget->setColumnHidden(FIELD_SEGMENTS_LIST, true);
}

DialogImportNZB::~DialogImportNZB()
{
    delete ui;
}

void DialogImportNZB::changeEvent(QEvent *e)
{
    QDialog::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

void DialogImportNZB::readNZBFile(QString filename)
{
    ui->editNZBFile->setText(filename);
    loadNZBFile();
    ui->btnAdd->setEnabled(true);
}

void DialogImportNZB::loadNZBFile(void)
{
    QFile file(ui->editNZBFile->text());

    if (!file.open(QFile::ReadOnly)) {
        qWarning("Failed to open NZB file");
        return;
    }

    QString nzb_data;

    ui->treeWidget->clear();
    nzbfile.files.clear();

    while (!file.atEnd()) {
        char buf[1024];

        if (file.readLine(buf, sizeof(buf)) == -1) {
            qWarning("Error occured while reading file: %s", file.errorString().toUtf8().data());
            file.close();
            return;
        }

        nzb_data.append(buf);
    }
    file.close();

    loadNZBFile(nzb_data);
}

void DialogImportNZB::loadNZBFile(QString nzb_data)
{
    QString errmsg;
    if (!glue->utilsParseNZBFile(nzb_data, nzbfile, errmsg)) {
        qWarning("Failed to import NZB file: %s", errmsg.toUtf8().data());
        return;
    }

    QNNTPGrabGlue::NZBFile file_in_nzb;

    Q_FOREACH (file_in_nzb, nzbfile.files) {
        QTreeWidgetItem *item = new QTreeWidgetItem();

        item->setCheckState(FIELD_DOWNLOAD, Qt::Checked);
        item->setText(FIELD_SUBJECT, file_in_nzb.subject);
        item->setText(FIELD_POSTER, file_in_nzb.poster);
        item->setData(FIELD_FILE_SIZE, Qt::DisplayRole, QVariant(file_in_nzb.file_size));
        item->setText(FIELD_FILE_SIZE_STR, glue->utilsCalculateFileSize(file_in_nzb.file_size));
        item->setText(FIELD_STAMP, file_in_nzb.stamp.toString());

        int j;
        QString groups;
        for (j = 0; j < file_in_nzb.groups.count(); j++) {
            if (groups.isEmpty()) {
                groups = file_in_nzb.groups[j];
            } else {
                groups += QString(",") + file_in_nzb.groups[j];
            }
        }
        item->setText(FIELD_GROUPS, groups);
        item->setData(FIELD_GROUPS_LIST, Qt::DisplayRole, QVariant::fromValue(file_in_nzb.groups));
        item->setData(FIELD_SEGMENTS_LIST, Qt::DisplayRole, QVariant::fromValue(file_in_nzb.segments));

        ui->treeWidget->addTopLevelItem(item);
    }

    ui->treeWidget->resizeColumnToContents(FIELD_SUBJECT);
    ui->treeWidget->resizeColumnToContents(FIELD_POSTER);
    ui->treeWidget->resizeColumnToContents(FIELD_GROUPS);
    ui->treeWidget->resizeColumnToContents(FIELD_STAMP);
}

void DialogImportNZB::btnSelectNZBFile_clicked(void)
{
    QString filename = QFileDialog::getOpenFileName(this, tr("Open NZB file"), NULL, tr("NZB Files (*.nzb);;All Files (*)"));
    if (filename.isEmpty()) {
        return;
    }

    ui->editNZBFile->setText(filename);

    loadNZBFile();

    ui->btnAdd->setEnabled(true);
}

void DialogImportNZB::mousePressEvent(QMouseEvent *event)
{
    if (event->type() == QEvent::MouseButtonPress && ui->editNZBFile->underMouse()) {
        btnSelectNZBFile_clicked();
    }
}

void DialogImportNZB::btnSelectAll_clicked(void)
{
    for (int i = 0; i < ui->treeWidget->topLevelItemCount(); i++) {
        QTreeWidgetItem *item = ui->treeWidget->topLevelItem(i);
        item->setCheckState(FIELD_DOWNLOAD, Qt::Checked);
    }
}

void DialogImportNZB::btnSelectNone_clicked(void)
{
    for (int i = 0; i < ui->treeWidget->topLevelItemCount(); i++) {
        QTreeWidgetItem *item = ui->treeWidget->topLevelItem(i);
        item->setCheckState(FIELD_DOWNLOAD, Qt::Unchecked);
    }
}

void DialogImportNZB::btnCancel_clicked(void)
{
    close();
}

void DialogImportNZB::btnAdd_clicked(void)
{
    QString collection_name;
    QFileInfo fi(ui->editNZBFile->text());

    collection_name = fi.baseName();

    DialogSetCollectionName d(glue, collection_name, this);
    switch (d.exec()) {
        case QDialog::Accepted:
            doImportNZB(d.getCollectionName());
            close();
            break;

        case QDialog::Rejected:
            break;
    }
}

void DialogImportNZB::doImportNZB(QString collection_name)
{
    for (int i = 0; i < ui->treeWidget->topLevelItemCount(); i++) {
        QTreeWidgetItem *item = ui->treeWidget->topLevelItem(i);

        /* Only import files which are checked */
        if (item->checkState(FIELD_DOWNLOAD) != Qt::Checked) {
            continue;
        }

        QString subject = item->data(FIELD_SUBJECT, Qt::DisplayRole).toString();
        QString poster = item->data(FIELD_POSTER, Qt::DisplayRole).toString();
        QDateTime stamp = QDateTime::fromString(item->data(FIELD_STAMP, Qt::DisplayRole).toString());
        quint64 file_size = item->data(FIELD_FILE_SIZE, Qt::DisplayRole).toLongLong();

        QList<QString> groups = item->data(FIELD_GROUPS_LIST, Qt::DisplayRole).value<QList<QString> >();
        QList<QNNTPGrabGlue::Part> segments = item->data(FIELD_SEGMENTS_LIST, Qt::DisplayRole).value<QList<QNNTPGrabGlue::Part> >();

        QString errmsg = "";

        if (!glue->schedularAddTaskToQueue(collection_name, subject, poster, stamp, file_size, groups, segments, errmsg)) {
            qWarning("%s", (tr("Unable to add file to the download queue: ") + errmsg).toUtf8().data());
        }
    }

    QString errmsg;
    if (!glue->schedularSaveQueue(errmsg)) {
        qWarning("%s", (tr("Unable to save download queue: ") + errmsg).toUtf8().data());
    }
}

void DialogImportNZB::quickImport(QString collection_name, QString nzb_data, bool allow_custom_collection_name)
{
    setVisible(false);
    loadNZBFile(nzb_data);
    if (allow_custom_collection_name) {
        DialogSetCollectionName d(glue, QNNTPGrabGlue::utilsSanitizeCollectionName(collection_name), this);
        switch (d.exec()) {
            case QDialog::Accepted:
                doImportNZB(d.getCollectionName());
                close();
                break;

            case QDialog::Rejected:
                break;
        }
    } else {
        doImportNZB(QNNTPGrabGlue::utilsSanitizeCollectionName(collection_name));
    }
}
