/*
    Copyright (C) 2005-2012  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>

#ifdef HAVE_LIBPROXY
#include <proxy.h>
#endif

#ifdef WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#include <wspiapi.h>
#include <io.h>
#undef  gai_strerror
#define gai_strerror gai_strerrorA
#define MSG_DONTWAIT 0
#else
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif

#include "download_thread.h"
#include "collection_alloc.h"
#include "nntpgrab_internal.h"
#include "download_queue.h"
#include "nntpgrab_plugin.h"

#include "nntpconnection.h"
#include "nntpconnection_backend_iface.h"
#include "nntpconnection_utils.h"

/* throttle.c */
void throttle_pause(struct timeval start_time, off_t xferlen, int max_bandwidth);

struct _NntpConnectionClass
{
    GObjectClass parent;
};

static void
nntpconnection_init (NntpConnection *obj)
{
}

static void
nntpconnection_class_init (NntpConnectionClass *klass)
{
}

G_DEFINE_TYPE(NntpConnection, nntpconnection, G_TYPE_OBJECT)

GType nntpconnection_backend_get_type(void)
{
    static GType object_type = 0;
    if (!object_type) {
        static const GTypeInfo object_info = {
            sizeof(NntpconnectionBackendIface),
            NULL,                               /* base init */
            NULL,                               /* base finalize */
        };

        object_type = g_type_register_static(G_TYPE_INTERFACE,
                                             "NntpconnectionBackend",
                                             &object_info, 0);
        }
    return object_type;
}

/************************
 * Wrapper functions
 ************************/
NNTPConnectionErrCode
nntpconnection_backend_connect(NntpConnection *conn, const char *hostname, int port, char **errmsg)
{
    g_return_val_if_fail(NNTPCONNECTION_BACKEND_GET_INTERFACE (conn)->connect, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);

    return (NNTPCONNECTION_BACKEND_GET_INTERFACE (conn)->connect(conn, hostname, port, errmsg));
}

void nntpconnection_backend_disconnect(NntpConnection *conn)
{
    g_return_if_fail(NNTPCONNECTION_BACKEND_GET_INTERFACE (conn)->disconnect);

    NNTPCONNECTION_BACKEND_GET_INTERFACE (conn)->disconnect (conn);
}

gboolean
nntpconnection_backend_prepare_ssl(NntpConnection *conn)
{
    char *errmsg = NULL;
    gboolean ret;

    g_return_val_if_fail(NNTPCONNECTION_BACKEND_GET_INTERFACE (conn)->prepare_ssl, FALSE);

    ret = (NNTPCONNECTION_BACKEND_GET_INTERFACE (conn)->prepare_ssl(conn, &errmsg));

    if (!ret) {
        /* Some error happened, disconnect the socket */
        g_return_val_if_fail(errmsg != NULL, FALSE);

        nntpconnection_disconnect_from_server(conn, DISCONNECT_ERROR_SSL_INITIALISE, errmsg, __FILE__, __LINE__);
        g_free(errmsg);

        return FALSE;
    }

    return TRUE;
}

gboolean
nntpconnection_ensure_data_is_ready(NntpConnection *conn, int timeout)
{
    g_return_val_if_fail(NNTPCONNECTION_BACKEND_GET_INTERFACE (conn)->ensure_data_is_ready, FALSE);

    return (NNTPCONNECTION_BACKEND_GET_INTERFACE (conn)->ensure_data_is_ready(conn, timeout));
}

gboolean
nntpconnection_read_msg(NntpConnection *conn, gboolean read_line, int max_length, void *data, int *data_length, gboolean *more_data_ready)
{
    gboolean try_again = FALSE;
    char *errmsg = NULL;
    gboolean ret;
    int data_length_tmp = 0;

    g_return_val_if_fail(NNTPCONNECTION_BACKEND_GET_INTERFACE (conn)->read_msg, FALSE);

    ret = (NNTPCONNECTION_BACKEND_GET_INTERFACE (conn)->read_msg(conn, read_line, max_length, data, &data_length_tmp, more_data_ready, &try_again, &errmsg));

    if (data_length) {
        *data_length = data_length_tmp;
    }

    if (!ret) {
        if (!try_again) {
            /* Some error happened, disconnect the socket */
            g_return_val_if_fail(errmsg != NULL, FALSE);

            nntpconnection_disconnect_from_server(conn, DISCONNECT_READ_ERROR, errmsg, __FILE__, __LINE__);
            g_free(errmsg);
        }

        return FALSE;
    }

    return TRUE;
}

gboolean
nntpconnection_send_msg(NntpConnection *conn, void *data, int length)
{
    char *errmsg = NULL;
    gboolean ret;

    g_return_val_if_fail(NNTPCONNECTION_BACKEND_GET_INTERFACE (conn)->send_msg, FALSE);

    if (length <= 0) {
        length = strlen(data);
    }

    ret = (NNTPCONNECTION_BACKEND_GET_INTERFACE (conn)->send_msg(conn, data, length, &errmsg));
    if (!ret) {
        /* Some error happened, disconnect the socket */
        g_return_val_if_fail(errmsg != NULL, FALSE);

        nntpconnection_disconnect_from_server(conn, DISCONNECT_WRITE_ERROR, errmsg, __FILE__, __LINE__);
        g_free(errmsg);

        return FALSE;
    }

    return TRUE;
}

gboolean nntpconnection_get_has_ssl_support(NntpConnection *conn)
{
    g_return_val_if_fail(NNTPCONNECTION_BACKEND_GET_INTERFACE (conn)->get_has_ssl_support, FALSE);

    return (NNTPCONNECTION_BACKEND_GET_INTERFACE (conn)->get_has_ssl_support(conn));
}

/*************************
 * Traffic monitoring
 *************************/

/* Keep track of the number of bytes received for the last 10 seconds */
static struct timeval tv_last_traffic_monitor_flush = { 0, 0 };
static volatile int bytes_received_in_last_sec;
static gboolean abort_traffic_thread = FALSE;

#ifdef WIN32
int gettimeofday (struct timeval *tv, void* tz)
{
  union {
    long long ns100; /*time since 1 Jan 1601 in 100ns units */
    FILETIME ft;
  } now;

  GetSystemTimeAsFileTime (&now.ft);
  tv->tv_usec = (long) ((now.ns100 / 10LL) % 1000000LL);
  tv->tv_sec = (long) ((now.ns100 - 116444736000000000LL) / 10000000LL);
  return (0);
}
#endif

gpointer
traffic_thread_func(gpointer data)
{
    int bytes_received[10];
    int empty_buf[10];
    gboolean null_buf_sent = FALSE;
    time_t timestamp = 0;
#if 0
    int i;
#endif

    memset(&bytes_received, 0, sizeof(bytes_received));
    memset(empty_buf, 0, sizeof(empty_buf));

    while (!abort_traffic_thread) {
        time_t prev_timestamp;

        // Traffic monitoring
        prev_timestamp = timestamp;
        timestamp = time(NULL);

        if (timestamp > prev_timestamp) {
            time_t diff;
            int bytes_received_copy[10];
            int val;

            // Calculate how many steps we need to shift
            diff = timestamp - prev_timestamp;
            if (diff > 10) {
                diff = 10;
            }

            /* This is not fully atomic, but will do for our goal */
            val = g_atomic_int_get(&bytes_received_in_last_sec);
            bytes_received[9] = g_atomic_int_exchange_and_add(&bytes_received_in_last_sec, -val);
            if (bytes_received[9] < 0) {
                bytes_received[9] = 0;
            }
            memcpy(bytes_received_copy, bytes_received, sizeof(bytes_received));

            // Shift all the previous values by 'diff'
            memmove(bytes_received, &bytes_received[diff], (10 - diff) * sizeof(int));
            memset(&bytes_received[10 - diff], 0, diff * sizeof(int));

            gettimeofday(&tv_last_traffic_monitor_flush, NULL);

            // Notify upstream about our values
            // If there hasn't been any sane data, send a empty buffer 1 time
            if (!memcmp(bytes_received_copy, empty_buf, sizeof(empty_buf))) {
                if (!null_buf_sent) {
                    null_buf_sent = TRUE;
                    nntpgrab_core_emit_traffic_monitor_update(FALSE, bytes_received_copy, timestamp);
                }
            } else {
                nntpgrab_core_emit_traffic_monitor_update(FALSE, bytes_received_copy, timestamp);
                null_buf_sent = FALSE;
            }

#if 0
            g_print("stamp = %li\n", timestamp);
            for (i = 0; i < 10; i++) {
                g_print("bytes_received[%i] = %i\n", i, bytes_received[i]);
            }
            g_print("\n");
#endif
        }

        g_usleep(G_USEC_PER_SEC * 1);
    }

    return NULL;
}

void
nntpconnection_backend_init_traffic_monitor(void)
{
    static gboolean traffic_monitor_started = FALSE;
    static GStaticMutex mutex = G_STATIC_MUTEX_INIT;

    g_static_mutex_lock(&mutex);
    if (!traffic_monitor_started) {
        traffic_monitor_started = TRUE;
        g_atomic_int_set(&bytes_received_in_last_sec, 0);

        /* TODO: implement a nice clean up */
        abort_traffic_thread = FALSE;
        /*traffic_thread = */g_thread_create(traffic_thread_func, NULL, TRUE, NULL);
    }
    g_static_mutex_unlock(&mutex);
}

void nntpconnection_backend_update_traffic_monitor(int bytes_read)
{
    g_atomic_int_add(&bytes_received_in_last_sec, bytes_read);
}

void nntpconnection_backend_perform_shaping(int max_bandwidth, int bytes_read)
{
    /* Perform traffic shaping if necessary */
    if (max_bandwidth > 0) {
        /* Calculate the sum of the number of bytes received in the last 10 seconds */
        int num_bytes_received = 0;

        num_bytes_received = g_atomic_int_get(&bytes_received_in_last_sec);
        num_bytes_received += bytes_read;
        throttle_pause(tv_last_traffic_monitor_flush, num_bytes_received, max_bandwidth);
    }
}

/************************************
 * Proxy settings
 ************************************/
static gboolean
get_proxy_settings(const char *hostname, char **proxy_host, int *proxy_port)
{
#ifdef HAVE_LIBPROXY
    static pxProxyFactory *proxy = NULL;
    static GStaticMutex mutex = G_STATIC_MUTEX_INIT;
    char hostname_with_https[256];
    char **proxies;
    char **values;
    int offset;

    g_return_val_if_fail(hostname != NULL, FALSE);
    g_return_val_if_fail(proxy_host != NULL, FALSE);
    g_return_val_if_fail(proxy_port != NULL, FALSE);

    /* initialize libproxy if it isn't already */
    g_static_mutex_lock(&mutex);
    if (proxy == NULL) {
        proxy = px_proxy_factory_new();
    }
    g_static_mutex_unlock(&mutex);

    /* libproxy requires the hostname to be prefixed with 'https://' */
    memset(&hostname_with_https, 0, sizeof(hostname_with_https));
    snprintf(hostname_with_https, sizeof(hostname_with_https) - 1, "https://%s", hostname);

    if (!proxy) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "Unable to retrieve proxy information, pxProxyFactory() returned NULL");
        return FALSE;
    }

    proxies = px_proxy_factory_get_proxies(proxy, hostname_with_https);
    if (!proxies) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "Unable to retrieve proxy information, px_proxy_factory_get_proxies() returned NULL");
        return FALSE;
    }

    if (!proxies[0]) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "Unable to retrieve proxy information, px_proxy_factory_get_proxies() returned an empty list");
        return FALSE;
    }

    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_DEBUG, "proxies[0] = %s", proxies[0]);

    if (!strcmp(proxies[0], "direct://")) {
        /* No proxy required */
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_DEBUG, "No proxy required for '%s'", hostname);
        return FALSE;
    }

    if (!strncmp(proxies[0], "http://", 7)) {
        offset = 7;
    } else if (!strncmp(proxies[0], "https://", 8)) {
        offset = 8;
    } else {
        offset = 0;
    }

    values = g_strsplit(proxies[0] + offset, ":", -1);
    if (!(values || values[0] || values[1])) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_WARNING, "Unable to retrieve proxy information, proxy '%s' could not be parsed", proxies[0]);
        return FALSE;
    }

    *proxy_host = g_strdup(values[0]);
    *proxy_port = atoi(values[1]);

    g_strfreev(values);

    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_DEBUG, "Returned proxy '%s:%i' for host '%s'", *proxy_host, *proxy_port, hostname);

    return TRUE;
#else
    /* No libproxy support, assume we always have a direct connection */
    return FALSE;
#endif
}

void
nntpconnection_init_libproxy(void)
{
    char *proxy_host = NULL;
    int proxy_port = 0;

    /* To avoid the situation where libproxy calls some X11 functions from a different
     * thread than the main loop we initialize libproxy here by performing a dummy query */
    if (get_proxy_settings("localhost", &proxy_host, &proxy_port)) {
        g_free(proxy_host);
    }
}

/********************/
/* proxy operations */
/********************/
void
nntpconnection_process_proxy_connect_command(NntpConnection *conn)
{
    char buf[128];
    int length = 0;
    char *ptr;

    g_return_if_fail(conn != NULL);

    memset(buf, 0, sizeof(buf));
    if (!nntpconnection_read_msg(conn, TRUE, sizeof(buf) - 1, buf, &length, NULL)) {
        return;
    }

    /* buf should now contain something like 'HTTP/1.0 200 Connection established' */
    ptr = strstr(buf, " ");
    if (nntpconnection_utils_get_status_code(ptr + 1) != 200) {
        /* Unknown response code, disconnect */
        nntpconnection_utils_strip_newline(buf, NULL);
        nntpconnection_disconnect_from_server(conn, DISCONNECT_INVALID_MSG, buf, __FILE__, __LINE__);
        return;
    }

    /* Keep on reading until we've found a blank line */
    do {
        memset(&buf, 0, sizeof(buf));
        if (!nntpconnection_read_msg(conn, TRUE, sizeof(buf) - 1, buf, &length, NULL)) {
            return;
        }

        nntpconnection_utils_strip_newline(buf, NULL);
    } while (strlen(buf) > 0);

    /* From now on we can talk the regular NNTP protocol */
    conn->active_method = nntpconnection_process_welcome_msg;

    /* Prepare the SSL connection if necessary */
    if (conn->server_info.use_ssl) {
        nntpconnection_backend_prepare_ssl(conn);
    }
}

NNTPConnectionErrCode
nntpconnection_connect_to_server(NntpConnection *conn, char **errmsg)
{
    char *proxy_host = NULL;
    int proxy_port = 0;
    NNTPConnectionErrCode retval;

    g_return_val_if_fail(conn != NULL, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(strlen(conn->server_info.hostname) > 0, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(conn->server_info.port > 0, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(errmsg != NULL, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(*errmsg == NULL, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);

    nntpconnection_backend_init_traffic_monitor();

    conn->is_idle = FALSE;
    conn->last_disconnect_stamp = time(NULL);       /* Assume that the connection attempt has failed by default */

    get_proxy_settings(conn->server_info.hostname, &proxy_host, &proxy_port);

    if (proxy_host) {
        retval = nntpconnection_backend_connect(conn, proxy_host, proxy_port, errmsg);
    } else {
        retval = nntpconnection_backend_connect(conn, conn->server_info.hostname, conn->server_info.port, errmsg);
    }

    if (retval != NNTP_CONNECTION_ERROR_NONE) {
        return retval;
    }

    /* Prepare a SSL connection if necessary (expect when we need to connect to a proxy server first) */
    if (conn->server_info.use_ssl && !proxy_host) {
        if (!nntpconnection_backend_prepare_ssl(conn)) {
            return NNTP_CONNECTION_ERROR_SSL_INITIALISE;
        }
    }

    /* Prepare connection via a proxy if necessary */
    if (proxy_host) {
        char cmd[128];

        memset(&cmd, 0, sizeof(cmd));
        snprintf(cmd, sizeof(cmd) - 1, "CONNECT %s:%i HTTP/1.1\r\nHost: %s:%i\r\n\r\n", conn->server_info.hostname, conn->server_info.port, conn->server_info.hostname, conn->server_info.port);
        nntpconnection_send_msg(conn, cmd, strlen(cmd));

        g_free(proxy_host);

        conn->active_method = nntpconnection_process_proxy_connect_command;
    } else {
        conn->active_method = nntpconnection_process_welcome_msg;
    }

    conn->last_activity_stamp = time(NULL);
    conn->last_disconnect_stamp = 0;

    return NNTP_CONNECTION_ERROR_NONE;
}

void
nntpconnection_disconnect_from_server(NntpConnection *conn, NNTPDisconnectType disconnect_type, const char *reason, const char *cause_file, int cause_lineno)
{
    g_return_if_fail(conn != NULL);
    g_return_if_fail(conn->poll_fd.fd >= 0);
    g_return_if_fail(cause_file != NULL);

    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "Request received to disconnect connection %i, poll_fd = %i, disconnect_type = %i, reason = %s, cause = %s:%i", conn->connection_id, (int) conn->poll_fd.fd, disconnect_type, reason, cause_file, cause_lineno);

    nntpconnection_backend_disconnect(conn);

    if (conn->collection && conn->file && conn->part) {
        /* Update the file status */
        download_queue_update_part_status(conn->server_info.servername, conn->connection_id, conn->collection, conn->file, conn->part, conn->server_id, FALSE, FALSE, FALSE);
    }

    nntpgrab_core_emit_disconnect(FALSE, conn->server_info.servername, conn->connection_id, disconnect_type, reason);

    conn->poll_fd.fd = -1;
    conn->is_idle = TRUE;
    if (disconnect_type == DISCONNECT_NORMAL) {
        conn->last_disconnect_stamp = 0;
    } else {
        conn->last_disconnect_stamp = time(NULL);
    }
    conn->last_activity_stamp = 0;

    if (conn->collection) {
        collection_unref(conn->collection);
        conn->collection = NULL;
    }

    if (conn->file) {
        file_unref(conn->file);
        conn->file = NULL;
    }

    if (conn->part) {
        conn->part = NULL;
    }

    conn->job_type = NNTP_JOB_TYPE_NONE;

    if (conn->article_fd >= 0) {
        close(conn->article_fd);
        conn->article_fd = -1;
    }
}
